@@ -1,9 +1,11 @@
 """
 Utilities for working with PDFs and redactions
 """
+import random
 import re
+import statistics
 import typing
-from typing import List, Tuple
+from typing import List
 
 import fitz
 from fitz import Page, Rect
@@ -11,6 +13,9 @@ from fitz import Page, Rect
 from .custom_types import CharDictType, RedactionType
 from .text_utils import is_ok_words, is_repeated_chars
 
+# Disable anti-aliasing when rendering and creating pixmaps
+fitz.TOOLS.set_aa_level(0)
+
 
 def get_good_rectangles(page: Page) -> List[Rect]:
     """Find rectangles in the PDFs that might be redactions.
@@ -30,15 +35,9 @@ def get_good_rectangles(page: Page) -> List[Rect]:
             # Not opaque. Probably a highlight or similar.
             continue
 
-        if drawing["fill"] in (
+        if drawing["fill"] is None:
             # Unfilled box (transparent to the eye, but distinct from ones that
             # have opacity of 0).
-            None,
-            # White
-            (1.0, 1.0, 1.0),
-        ):
-            # White or unfilled box. These are used for various purposes
-            # like, with line number columns, borders, etc. Ignore them.
             continue
 
         # Each drawing can contain multiple "draw" commands that could be
@@ -67,7 +66,7 @@ def get_good_rectangles(page: Page) -> List[Rect]:
         for rectangle in rectangles:
             # Give it the sequence number and color of its parent drawing
             rectangle.seqno = drawing["seqno"]
-            rectangle.color = drawing["fill"]
+            rectangle.fill = drawing["fill"]
             if rectangle.y1 <= 43:
                 # It's a header, ignore it
                 continue
@@ -94,7 +93,7 @@ def intersects(
     """Determine if a rectangle intersects is occluded by a list of others
 
     This uses Rect objects, but note that they must have extra attributes of
-    "color" and "seqno".
+    "fill" and "seqno".
 
     :param text_rect: The rectangle around the text to check for intersections.
     :param rectangles: A list of rectangles to check for intersections.
@@ -108,19 +107,19 @@ def intersects(
     """
     for rect in rectangles + [text_rect]:
         assert all(
-            [hasattr(rect, "seqno"), hasattr(rect, "color")]
-        ), "Rectangle lacks required 'seqno' or 'color' attribute."
+            [hasattr(rect, "seqno"), hasattr(rect, "fill")]
+        ), "Rectangle lacks required 'seqno' or 'fill' attribute."
 
     overlapping_areas = []
     for rect in rectangles:
         intersecting_area = abs(text_rect & rect)
         if intersecting_area > 0 and rect.seqno > text_rect.seqno:
-            # Intersecting and text was drawn first,
-            # meaning it's behind the drawing.
+            # Intersecting text was drawn first, meaning it's behind the rect.
             overlapping_areas.append(intersecting_area)
             continue
-        if intersecting_area > 0 and rect.color == text_rect.color:
-            # Intersecting and same color. This makes text invisible.
+        if intersecting_area > 0 and rect.fill == text_rect.fill:
+            # Intersecting and same color. This makes text invisible even if
+            # it's drawn on top of the rect.
             overlapping_areas.append(intersecting_area)
             continue
 
@@ -159,13 +158,13 @@ def get_intersecting_chars(
         span_color = span["color"]
         span_rect = fitz.Rect(span["bbox"])
         span_rect.seqno = span_seq_no
-        span_rect.color = span_color
+        span_rect.fill = span_color
         if not intersects(span_rect, rectangles):
             continue
         for char in span["chars"]:
             char_rect = fitz.Rect(char[3])
             char_rect.seqno = span_seq_no
-            char_rect.color = span_color
+            char_rect.fill = span_color
             if intersects(char_rect, rectangles, occlusion_threshold=0.8):
                 char_dict: CharDictType = {
                     "rect": char_rect,
@@ -184,6 +183,9 @@ def group_chars_by_rect(
     intersected with and group the chars back into words along with the bboxes
     of the rectangles they intersected with.
 
+    If a char intersects with more than one rectangle, only include it as part
+    of the rectangle with the highest sequence number.
+
     :param chars: The list of character dicts that intersect with rectangles in
     the PDF.
     :param rectangles: A list of PyMuPDF.Rect objects from the page (aka the
@@ -191,22 +193,46 @@ def group_chars_by_rect(
     :return: A list of dictionaries with keys for the rectangle's BBOX and the
     words underneath it.
     """
+    # A problem that we must deal with is stacked rectangles. Imagine a stack
+    # of stuff like so:
+    #
+    #   On top: Some red characters, "ABC"
+    #   Then: A white rect
+    #   On bottom: A red rect
+    #
+    # In this case, you can see the letters because they have a white
+    # background. It's not a bad redaction, even though the letters intersect
+    # with each of the rectangles. These need to get coalesced into a single
+    # bad redaction.
+    #
+    # Reverse-sort the rectangles by sequence number, and eliminate each char
+    # as soon as it intersects a rectangle.
     redactions = []
-    for rect in rectangles:
+    # Sort the rectangles by reversed sequence key.
+    seq_sorted_rects = sorted(rectangles, key=lambda x: x.seqno, reverse=True)
+    for rect in seq_sorted_rects:
         redaction: RedactionType = {
             "bbox": (rect.x0, rect.y0, rect.x1, rect.y1),
             "text": "",
         }
-        for char in chars:
-            if char["rect"] & rect:
+        # Make a copy of the chars list so we can manipulate it in the loop
+        char_copy = chars.copy()
+        for char in char_copy:
+            if abs(char["rect"] & rect):
+                # The char intersects with this rectangle. Add it to the
+                # redaction dict, and remove it from the list so it doesn't
+                # get analyzed again.
                 redaction["text"] += char["c"]
+                chars.remove(char)
         redactions.append(redaction)
 
     return redactions
 
 
 @typing.no_type_check  # It gets confused w/filters
-def filter_redactions(redactions: List[RedactionType]) -> List[RedactionType]:
+def filter_redactions_by_text(
+    redactions: List[RedactionType],
+) -> List[RedactionType]:
     """Filter out redactions that are not actually bad.
 
     :param redactions: A list of redactions that might be bad
@@ -218,7 +244,8 @@ def filter_redactions(redactions: List[RedactionType]) -> List[RedactionType]:
     # Has non-whitespace content and isn't blank
     redactions = filter(lambda r: r["text"].strip(), redactions)
 
-    # Has some letters or numbers
+    # Has some letters or numbers (isn't just a rectangle
+    # with nothing under it)
     redactions = filter(lambda r: re.search(r"[\d\w]", r["text"]), redactions)
 
     # Has OK words in redaction
@@ -227,6 +254,42 @@ def filter_redactions(redactions: List[RedactionType]) -> List[RedactionType]:
     return list(redactions)
 
 
+def filter_redactions_by_pixmap(
+    redactions: List[RedactionType],
+    page: Page,
+) -> List[RedactionType]:
+    """Convert each bad redaction to an image and check it for text
+
+    :param redactions: A list of redactions that might be bad
+    :param page: The PyMuPDF.Page object where the bad redactions might be
+    :return: The redactions, if they are valid
+    """
+    bad_redactions = []
+    for redaction in redactions:
+        pixmap = page.get_pixmap(
+            # Use gray for simplicity and speed, though this risks missing a
+            # bad redaction.
+            colorspace=fitz.csGRAY,
+            clip=fitz.Rect(redaction["bbox"]),
+        )
+        # Get a memoryview of the pixels. This is a 1-D array of grayscale
+        # values.
+        pixels = pixmap.samples_mv
+        if len(pixels) > 1000:
+            # Big redaction. Use random sampling to select fewer pixels
+            pixels = random.sample(pixels, 1000)
+        std_dev = statistics.stdev(pixels)
+        if std_dev > 0:
+            # There's some degree of variation in the grayscale values of the
+            # pixels. ∴ it's not a uniform box and it's not a bad redaction.
+            # filename = f'{redaction["text"].replace("/", "_")}.png'
+            # print(f"Got high variance of {std_dev}. Saving {filename}")
+            # pixmap.save(filename)
+            continue
+        bad_redactions.append(redaction)
+    return bad_redactions
+
+
 def get_bad_redactions(page: Page) -> List[RedactionType]:
     """Get the bad redactions for a page from a PDF
 
@@ -237,5 +300,6 @@ def get_bad_redactions(page: Page) -> List[RedactionType]:
     good_rectangles = get_good_rectangles(page)
     intersecting_chars = get_intersecting_chars(page, good_rectangles)
     redactions = group_chars_by_rect(intersecting_chars, good_rectangles)
-    bad_redactions = filter_redactions(redactions)
+    bad_redactions = filter_redactions_by_text(redactions)
+    bad_redactions = filter_redactions_by_pixmap(bad_redactions, page)
     return bad_redactions
